@@ -60,17 +60,25 @@ typedef struct {
 
 static p11_local * _p11_library_get_thread_local (void);
 
+#ifdef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+p11_mutex_t p11_library_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
+
+p11_mutex_t p11_virtual_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
+#else
 p11_mutex_t p11_library_mutex;
 
 p11_mutex_t p11_virtual_mutex;
+#endif
 
 #ifdef OS_UNIX
+#ifndef __GNUC__
 pthread_once_t p11_library_once = PTHREAD_ONCE_INIT;
 #endif
+#endif
 
 unsigned int p11_forkid = 1;
 
-#ifdef HAVE_LOCALE_H
+#ifdef HAVE_STRERROR_L
 extern locale_t p11_message_locale;
 #endif
 
@@ -96,6 +104,21 @@ uninit_common (void)
 
 #ifdef OS_UNIX
 
+#ifdef P11_TLS_KEYWORD
+static p11_local *
+_p11_library_get_thread_local (void)
+{
+	static P11_TLS_KEYWORD p11_local local;
+	static P11_TLS_KEYWORD bool local_initialized = false;
+
+	if (!local_initialized) {
+		memset (&local, 0, sizeof (p11_local));
+		local_initialized = true;
+	}
+
+	return &local;
+}
+#else
 static pthread_key_t thread_local = 0;
 
 static p11_local *
@@ -113,6 +136,26 @@ _p11_library_get_thread_local (void)
 
 	return local;
 }
+#endif
+
+#if defined(HAVE_DECL___REGISTER_ATFORK) && !HAVE_DECL___REGISTER_ATFORK
+extern int __register_atfork (void (*prepare) (void), void (*parent) (void),
+			      void (*child) (void), void *dso_handle);
+#endif /* HAVE_DECL___REGISTER_ATFORK */
+
+#ifdef HAVE___REGISTER_ATFORK
+
+extern void *__dso_handle;
+
+#define p11_register_atfork(a,b,c) \
+	(__register_atfork((a),(b),(c),__dso_handle))
+
+#else
+
+#define p11_register_atfork(a,b,c) \
+	(pthread_atfork((a),(b),(c)))
+
+#endif /* HAVE___REGISTER_ATFORK */
 
 static void
 count_forks (void)
@@ -126,21 +169,23 @@ p11_library_init_impl (void)
 {
 	p11_debug_init ();
 	p11_debug ("initializing library");
-	p11_mutex_init (&p11_library_mutex);
-	p11_mutex_init (&p11_virtual_mutex);
+	P11_RECURSIVE_MUTEX_INIT (p11_library_mutex);
+	P11_RECURSIVE_MUTEX_INIT (p11_virtual_mutex);
+#ifndef P11_TLS_KEYWORD
 	pthread_key_create (&thread_local, free);
+#endif
 	p11_message_storage = thread_local_message;
-#ifdef HAVE_NEWLOCALE
+#ifdef HAVE_STRERROR_L
 	p11_message_locale = newlocale (LC_ALL_MASK, "POSIX", (locale_t) 0);
 #endif
 
-	pthread_atfork (NULL, NULL, count_forks);
+	p11_register_atfork (NULL, NULL, count_forks);
 }
 
 void
 p11_library_init (void)
 {
-	p11_library_init_once ();
+	p11_library_init_impl ();
 }
 
 void
@@ -148,15 +193,19 @@ p11_library_uninit (void)
 {
 	uninit_common ();
 
+#ifndef P11_TLS_KEYWORD
 	/* Some cleanup to pacify valgrind */
 	free (pthread_getspecific (thread_local));
 	pthread_setspecific (thread_local, NULL);
+#endif
 
-#ifdef HAVE_NEWLOCALE
+#ifdef HAVE_STRERROR_L
 	freelocale (p11_message_locale);
 #endif
 	p11_message_storage = dont_store_message;
+#ifndef P11_TLS_KEYWORD
 	pthread_key_delete (thread_local);
+#endif
 	p11_mutex_uninit (&p11_virtual_mutex);
 	p11_mutex_uninit (&p11_library_mutex);
 }
@@ -191,8 +240,8 @@ p11_library_init (void)
 {
 	p11_debug_init ();
 	p11_debug ("initializing library");
-	p11_mutex_init (&p11_library_mutex);
-	p11_mutex_init (&p11_virtual_mutex);
+	P11_RECURSIVE_MUTEX_INIT (p11_library_mutex);
+	P11_RECURSIVE_MUTEX_INIT (p11_virtual_mutex);
 	thread_local = TlsAlloc ();
 	if (thread_local == TLS_OUT_OF_INDEXES)
 		p11_debug ("couldn't setup tls");
