@@ -88,14 +88,14 @@ typedef struct {
 typedef struct _State {
 	p11_virtual virt;
 	struct _State *next;
+	CK_FUNCTION_LIST **loaded;
 	CK_FUNCTION_LIST *wrapped;
 	CK_ULONG last_handle;
 	Proxy *px;
 } State;
 
-static CK_FUNCTION_LIST **all_modules = NULL;
 static State *all_instances = NULL;
-static State global = { { { { -1, -1 }, NULL, }, }, NULL, NULL, FIRST_HANDLE, NULL };
+static State global = { { { { -1, -1 }, NULL, }, }, NULL, NULL, NULL, FIRST_HANDLE, NULL };
 
 #define PROXY_VALID(px) ((px) && (px)->forkid == p11_forkid)
 #define PROXY_FORKED(px) ((px) && (px)->forkid != p11_forkid)
@@ -123,6 +123,8 @@ map_slot_unlocked (Proxy *px,
 
 	if (slot > px->n_mappings) {
 		return CKR_SLOT_ID_INVALID;
+	} else if (px->n_mappings == 0) {
+		return CKR_SLOT_ID_INVALID;
 	} else {
 		assert (px->mappings);
 		memcpy (mapping, &px->mappings[slot], sizeof (Mapping));
@@ -248,12 +250,14 @@ modules_dup (CK_FUNCTION_LIST **modules)
 }
 
 static CK_RV
-proxy_create (Proxy **res)
+proxy_create (Proxy **res, CK_FUNCTION_LIST **loaded,
+	      Mapping *mappings, unsigned int n_mappings)
 {
 	CK_FUNCTION_LIST_PTR *f;
 	CK_FUNCTION_LIST_PTR funcs;
 	CK_SLOT_ID_PTR slots;
 	CK_ULONG i, count;
+	unsigned int j;
 	CK_RV rv = CKR_OK;
 	Proxy *py;
 
@@ -262,7 +266,7 @@ proxy_create (Proxy **res)
 
 	py->forkid = p11_forkid;
 
-	py->inited = modules_dup (all_modules);
+	py->inited = modules_dup (loaded);
 	return_val_if_fail (py->inited != NULL, CKR_HOST_MEMORY);
 
 	rv = p11_kit_modules_initialize (py->inited, NULL);
@@ -293,8 +297,14 @@ proxy_create (Proxy **res)
 
 				/* And now add a mapping for each of those slots */
 				for (i = 0; i < count; ++i) {
+					/* Reuse the existing mapping if any */
+					for (j = 0; j < n_mappings; ++j) {
+						if (mappings[j].funcs == funcs &&
+						    mappings[j].real_slot == slots[i])
+							break;
+					}
 					py->mappings[py->n_mappings].funcs = funcs;
-					py->mappings[py->n_mappings].wrap_slot = py->n_mappings + MAPPING_OFFSET;
+					py->mappings[py->n_mappings].wrap_slot = j == n_mappings ? py->n_mappings + MAPPING_OFFSET : mappings[j].wrap_slot;
 					py->mappings[py->n_mappings].real_slot = slots[i];
 					++py->n_mappings;
 				}
@@ -323,6 +333,8 @@ proxy_C_Initialize (CK_X_FUNCTION_LIST *self,
 {
 	State *state = (State *)self;
 	bool initialize = false;
+	Mapping *mappings = NULL;
+	unsigned int n_mappings = 0;
 	Proxy *py;
 	CK_RV rv;
 
@@ -338,8 +350,15 @@ proxy_C_Initialize (CK_X_FUNCTION_LIST *self,
 			unsigned call_finalize = 1;
 
 			initialize = true;
-			if (PROXY_FORKED(state->px))
+			if (PROXY_FORKED(state->px)) {
 				call_finalize = 0;
+				if (state->px->mappings) {
+					mappings = state->px->mappings;
+					n_mappings = state->px->n_mappings;
+					state->px->mappings = NULL;
+					state->px->n_mappings = 0;
+				}
+			}
 			proxy_free (state->px, call_finalize);
 
 			state->px = NULL;
@@ -354,7 +373,8 @@ proxy_C_Initialize (CK_X_FUNCTION_LIST *self,
 		return CKR_OK;
 	}
 
-	rv = proxy_create (&py);
+	rv = proxy_create (&py, state->loaded, mappings, n_mappings);
+	free (mappings);
 	if (rv != CKR_OK) {
 		p11_debug ("out: %lu", rv);
 		return rv;
@@ -401,8 +421,8 @@ proxy_C_GetInfo (CK_X_FUNCTION_LIST *self,
 	info->libraryVersion.major = LIBRARY_VERSION_MAJOR;
 	info->libraryVersion.minor = LIBRARY_VERSION_MINOR;
 	info->flags = 0;
-	strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
-	strncpy ((char*)info->libraryDescription, LIBRARY_DESCRIPTION, 32);
+	memcpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
+	memcpy ((char*)info->libraryDescription, LIBRARY_DESCRIPTION, 32);
 	return CKR_OK;
 }
 
@@ -665,34 +685,6 @@ proxy_C_CloseAllSessions (CK_X_FUNCTION_LIST *self,
 	return CKR_OK;
 }
 
-static CK_RV
-proxy_C_GetFunctionStatus (CK_X_FUNCTION_LIST *self,
-                           CK_SESSION_HANDLE handle)
-{
-	State *state = (State *)self;
-	Mapping map;
-	CK_RV rv;
-
-	rv = map_session_to_real (state->px, &handle, &map, NULL);
-	if (rv != CKR_OK)
-		return rv;
-	return (map.funcs->C_GetFunctionStatus) (handle);
-}
-
-static CK_RV
-proxy_C_CancelFunction (CK_X_FUNCTION_LIST *self,
-                        CK_SESSION_HANDLE handle)
-{
-	State *state = (State *)self;
-	Mapping map;
-	CK_RV rv;
-
-	rv = map_session_to_real (state->px, &handle, &map, NULL);
-	if (rv != CKR_OK)
-		return rv;
-	return (map.funcs->C_CancelFunction) (handle);
-}
-
 static CK_RV
 proxy_C_GetSessionInfo (CK_X_FUNCTION_LIST *self,
                         CK_SESSION_HANDLE handle,
@@ -1584,85 +1576,6 @@ proxy_C_GenerateRandom (CK_X_FUNCTION_LIST *self,
 	return (map.funcs->C_GenerateRandom) (handle, random_data, random_len);
 }
 
-/* --------------------------------------------------------------------
- * Global module functions
- */
-
-static CK_FUNCTION_LIST module_functions;
-
-static CK_RV
-module_C_Initialize (CK_VOID_PTR init_args)
-{
-	return proxy_C_Initialize (&global.virt.funcs, init_args);
-}
-
-static CK_RV
-module_C_Finalize (CK_VOID_PTR reserved)
-{
-	return proxy_C_Finalize (&global.virt.funcs, reserved);
-}
-
-static CK_RV
-module_C_GetInfo (CK_INFO_PTR info)
-{
-	return proxy_C_GetInfo (&global.virt.funcs, info);
-}
-
-static CK_RV
-module_C_GetFunctionList (CK_FUNCTION_LIST_PTR_PTR list)
-{
-	return_val_if_fail (list != NULL, CKR_ARGUMENTS_BAD);
-	*list = &module_functions;
-	return CKR_OK;
-}
-
-static CK_RV
-module_C_GetSlotList (CK_BBOOL token_present,
-                      CK_SLOT_ID_PTR slot_list,
-                      CK_ULONG_PTR count)
-{
-	return proxy_C_GetSlotList (&global.virt.funcs, token_present, slot_list, count);
-}
-
-static CK_RV
-module_C_GetSlotInfo (CK_SLOT_ID id,
-                      CK_SLOT_INFO_PTR info)
-{
-	return proxy_C_GetSlotInfo (&global.virt.funcs, id, info);
-}
-
-static CK_RV
-module_C_GetTokenInfo (CK_SLOT_ID id,
-                       CK_TOKEN_INFO_PTR info)
-{
-	return proxy_C_GetTokenInfo (&global.virt.funcs, id, info);
-}
-
-static CK_RV
-module_C_GetMechanismList (CK_SLOT_ID id,
-                           CK_MECHANISM_TYPE_PTR mechanism_list,
-                           CK_ULONG_PTR count)
-{
-	return proxy_C_GetMechanismList (&global.virt.funcs, id, mechanism_list, count);
-}
-
-static CK_RV
-module_C_GetMechanismInfo (CK_SLOT_ID id,
-                           CK_MECHANISM_TYPE type,
-                           CK_MECHANISM_INFO_PTR info)
-{
-	return proxy_C_GetMechanismInfo (&global.virt.funcs, id, type, info);
-}
-
-static CK_RV
-module_C_InitToken (CK_SLOT_ID id,
-                    CK_UTF8CHAR_PTR pin,
-                    CK_ULONG pin_len,
-                    CK_UTF8CHAR_PTR label)
-{
-	return proxy_C_InitToken (&global.virt.funcs, id, pin, pin_len, label);
-}
-
 static CK_RV
 module_C_WaitForSlotEvent (CK_FLAGS flags,
                            CK_SLOT_ID_PTR slot,
@@ -1671,608 +1584,10 @@ module_C_WaitForSlotEvent (CK_FLAGS flags,
 	return proxy_C_WaitForSlotEvent (&global.virt.funcs, flags, slot, reserved);
 }
 
-static CK_RV
-module_C_OpenSession (CK_SLOT_ID id,
-                      CK_FLAGS flags,
-                      CK_VOID_PTR user_data,
-                      CK_NOTIFY callback,
-                      CK_SESSION_HANDLE_PTR handle)
-{
-	return proxy_C_OpenSession (&global.virt.funcs, id, flags, user_data, callback,
-	                            handle);
-}
-
-static CK_RV
-module_C_CloseSession (CK_SESSION_HANDLE handle)
-{
-	return proxy_C_CloseSession (&global.virt.funcs, handle);
-}
-
-static CK_RV
-module_C_CloseAllSessions (CK_SLOT_ID id)
-{
-	return proxy_C_CloseAllSessions (&global.virt.funcs, id);
-}
-
-static CK_RV
-module_C_GetFunctionStatus (CK_SESSION_HANDLE handle)
-{
-	return proxy_C_GetFunctionStatus (&global.virt.funcs, handle);
-}
-
-static CK_RV
-module_C_CancelFunction (CK_SESSION_HANDLE handle)
-{
-	return proxy_C_CancelFunction (&global.virt.funcs, handle);
-}
-
-static CK_RV
-module_C_GetSessionInfo (CK_SESSION_HANDLE handle,
-                         CK_SESSION_INFO_PTR info)
-{
-	return proxy_C_GetSessionInfo (&global.virt.funcs, handle, info);
-}
-
-static CK_RV
-module_C_InitPIN (CK_SESSION_HANDLE handle,
-                  CK_UTF8CHAR_PTR pin,
-                  CK_ULONG pin_len)
-{
-	return proxy_C_InitPIN (&global.virt.funcs, handle, pin, pin_len);
-}
-
-static CK_RV
-module_C_SetPIN (CK_SESSION_HANDLE handle,
-                 CK_UTF8CHAR_PTR old_pin,
-                 CK_ULONG old_pin_len,
-                 CK_UTF8CHAR_PTR new_pin,
-                 CK_ULONG new_pin_len)
-{
-	return proxy_C_SetPIN (&global.virt.funcs, handle, old_pin, old_pin_len, new_pin,
-	                       new_pin_len);
-}
-
-static CK_RV
-module_C_GetOperationState (CK_SESSION_HANDLE handle,
-                            CK_BYTE_PTR operation_state,
-                            CK_ULONG_PTR operation_state_len)
-{
-	return proxy_C_GetOperationState (&global.virt.funcs, handle, operation_state,
-	                                  operation_state_len);
-}
-
-static CK_RV
-module_C_SetOperationState (CK_SESSION_HANDLE handle,
-                            CK_BYTE_PTR operation_state,
-                            CK_ULONG operation_state_len,
-                            CK_OBJECT_HANDLE encryption_key,
-                            CK_OBJECT_HANDLE authentication_key)
-{
-	return proxy_C_SetOperationState (&global.virt.funcs, handle, operation_state,
-	                                  operation_state_len, encryption_key,
-	                                  authentication_key);
-}
-
-static CK_RV
-module_C_Login (CK_SESSION_HANDLE handle,
-                CK_USER_TYPE user_type,
-                CK_UTF8CHAR_PTR pin,
-                CK_ULONG pin_len)
-{
-	return proxy_C_Login (&global.virt.funcs, handle, user_type, pin, pin_len);
-}
-
-static CK_RV
-module_C_Logout (CK_SESSION_HANDLE handle)
-{
-	return proxy_C_Logout (&global.virt.funcs, handle);
-}
-
-static CK_RV
-module_C_CreateObject (CK_SESSION_HANDLE handle,
-                       CK_ATTRIBUTE_PTR template,
-                       CK_ULONG count,
-                       CK_OBJECT_HANDLE_PTR new_object)
-{
-	return proxy_C_CreateObject (&global.virt.funcs, handle, template, count,
-	                             new_object);
-}
-
-static CK_RV
-module_C_CopyObject (CK_SESSION_HANDLE handle,
-                     CK_OBJECT_HANDLE object,
-                     CK_ATTRIBUTE_PTR template,
-                     CK_ULONG count,
-                     CK_OBJECT_HANDLE_PTR new_object)
-{
-	return proxy_C_CopyObject (&global.virt.funcs, handle, object, template, count,
-	                           new_object);
-}
-
-static CK_RV
-module_C_DestroyObject (CK_SESSION_HANDLE handle,
-                        CK_OBJECT_HANDLE object)
-{
-	return proxy_C_DestroyObject (&global.virt.funcs, handle, object);
-}
-
-static CK_RV
-module_C_GetObjectSize (CK_SESSION_HANDLE handle,
-                        CK_OBJECT_HANDLE object,
-                        CK_ULONG_PTR size)
-{
-	return proxy_C_GetObjectSize (&global.virt.funcs, handle, object, size);
-}
-
-static CK_RV
-module_C_GetAttributeValue (CK_SESSION_HANDLE handle,
-                            CK_OBJECT_HANDLE object,
-                            CK_ATTRIBUTE_PTR template,
-                            CK_ULONG count)
-{
-	return proxy_C_GetAttributeValue (&global.virt.funcs, handle, object, template,
-	                                  count);
-}
-
-static CK_RV
-module_C_SetAttributeValue (CK_SESSION_HANDLE handle,
-                            CK_OBJECT_HANDLE object,
-                            CK_ATTRIBUTE_PTR template,
-                            CK_ULONG count)
-{
-	return proxy_C_SetAttributeValue (&global.virt.funcs, handle, object, template,
-	                                  count);
-}
-
-static CK_RV
-module_C_FindObjectsInit (CK_SESSION_HANDLE handle,
-                          CK_ATTRIBUTE_PTR template,
-                          CK_ULONG count)
-{
-	return proxy_C_FindObjectsInit (&global.virt.funcs, handle, template, count);
-}
-
-static CK_RV
-module_C_FindObjects (CK_SESSION_HANDLE handle,
-                      CK_OBJECT_HANDLE_PTR objects,
-                      CK_ULONG max_count,
-                      CK_ULONG_PTR count)
-{
-	return proxy_C_FindObjects (&global.virt.funcs, handle, objects, max_count, count);
-}
-
-static CK_RV
-module_C_FindObjectsFinal (CK_SESSION_HANDLE handle)
-{
-	return proxy_C_FindObjectsFinal (&global.virt.funcs, handle);
-}
-
-static CK_RV
-module_C_EncryptInit (CK_SESSION_HANDLE handle,
-                      CK_MECHANISM_PTR mechanism,
-                      CK_OBJECT_HANDLE key)
-{
-	return proxy_C_EncryptInit (&global.virt.funcs, handle, mechanism, key);
-}
-
-static CK_RV
-module_C_Encrypt (CK_SESSION_HANDLE handle,
-                  CK_BYTE_PTR data,
-                  CK_ULONG data_len,
-                  CK_BYTE_PTR encrypted_data,
-                  CK_ULONG_PTR encrypted_data_len)
-{
-	return proxy_C_Encrypt (&global.virt.funcs, handle, data, data_len,
-	                        encrypted_data, encrypted_data_len);
-}
-
-static CK_RV
-module_C_EncryptUpdate (CK_SESSION_HANDLE handle,
-                        CK_BYTE_PTR part,
-                        CK_ULONG part_len,
-                        CK_BYTE_PTR encrypted_part,
-                        CK_ULONG_PTR encrypted_part_len)
-{
-	return proxy_C_EncryptUpdate (&global.virt.funcs, handle, part, part_len,
-	                              encrypted_part, encrypted_part_len);
-}
-
-static CK_RV
-module_C_EncryptFinal (CK_SESSION_HANDLE handle,
-                       CK_BYTE_PTR last_part,
-                       CK_ULONG_PTR last_part_len)
-{
-	return proxy_C_EncryptFinal (&global.virt.funcs, handle, last_part, last_part_len);
-}
-
-static CK_RV
-module_C_DecryptInit (CK_SESSION_HANDLE handle,
-                      CK_MECHANISM_PTR mechanism,
-                      CK_OBJECT_HANDLE key)
-{
-	return proxy_C_DecryptInit (&global.virt.funcs, handle, mechanism, key);
-}
-
-static CK_RV
-module_C_Decrypt (CK_SESSION_HANDLE handle,
-                  CK_BYTE_PTR enc_data,
-                  CK_ULONG enc_data_len,
-                  CK_BYTE_PTR data,
-                  CK_ULONG_PTR data_len)
-{
-	return proxy_C_Decrypt (&global.virt.funcs, handle, enc_data, enc_data_len,
-	                        data, data_len);
-}
-
-static CK_RV
-module_C_DecryptUpdate (CK_SESSION_HANDLE handle,
-                        CK_BYTE_PTR enc_part,
-                        CK_ULONG enc_part_len,
-                        CK_BYTE_PTR part,
-                        CK_ULONG_PTR part_len)
-{
-	return proxy_C_DecryptUpdate (&global.virt.funcs, handle, enc_part, enc_part_len,
-	                              part, part_len);
-}
-
-static CK_RV
-module_C_DecryptFinal (CK_SESSION_HANDLE handle,
-                       CK_BYTE_PTR last_part,
-                       CK_ULONG_PTR last_part_len)
-{
-	return proxy_C_DecryptFinal (&global.virt.funcs, handle, last_part, last_part_len);
-}
-
-static CK_RV
-module_C_DigestInit (CK_SESSION_HANDLE handle,
-                     CK_MECHANISM_PTR mechanism)
-{
-	return proxy_C_DigestInit (&global.virt.funcs, handle, mechanism);
-}
-
-static CK_RV
-module_C_Digest (CK_SESSION_HANDLE handle,
-                 CK_BYTE_PTR data,
-                 CK_ULONG data_len,
-                 CK_BYTE_PTR digest,
-                 CK_ULONG_PTR digest_len)
-{
-	return proxy_C_Digest (&global.virt.funcs, handle, data, data_len, digest,
-	                       digest_len);
-}
-
-static CK_RV
-module_C_DigestUpdate (CK_SESSION_HANDLE handle,
-                       CK_BYTE_PTR part,
-                       CK_ULONG part_len)
-{
-	return proxy_C_DigestUpdate (&global.virt.funcs, handle, part, part_len);
-}
-
-static CK_RV
-module_C_DigestKey (CK_SESSION_HANDLE handle,
-                    CK_OBJECT_HANDLE key)
-{
-	return proxy_C_DigestKey (&global.virt.funcs, handle, key);
-}
-
-static CK_RV
-module_C_DigestFinal (CK_SESSION_HANDLE handle,
-                      CK_BYTE_PTR digest,
-                      CK_ULONG_PTR digest_len)
-{
-	return proxy_C_DigestFinal (&global.virt.funcs, handle, digest, digest_len);
-}
-
-static CK_RV
-module_C_SignInit (CK_SESSION_HANDLE handle,
-                   CK_MECHANISM_PTR mechanism,
-                   CK_OBJECT_HANDLE key)
-{
-	return proxy_C_SignInit (&global.virt.funcs, handle, mechanism, key);
-}
-
-static CK_RV
-module_C_Sign (CK_SESSION_HANDLE handle,
-               CK_BYTE_PTR data,
-               CK_ULONG data_len,
-               CK_BYTE_PTR signature,
-               CK_ULONG_PTR signature_len)
-{
-	return proxy_C_Sign (&global.virt.funcs, handle, data, data_len, signature,
-	                     signature_len);
-}
-
-static CK_RV
-module_C_SignUpdate (CK_SESSION_HANDLE handle,
-                     CK_BYTE_PTR part,
-                     CK_ULONG part_len)
-{
-	return proxy_C_SignUpdate (&global.virt.funcs, handle, part, part_len);
-}
-
-static CK_RV
-module_C_SignFinal (CK_SESSION_HANDLE handle,
-                    CK_BYTE_PTR signature,
-                    CK_ULONG_PTR signature_len)
-{
-	return proxy_C_SignFinal (&global.virt.funcs, handle, signature, signature_len);
-}
-
-static CK_RV
-module_C_SignRecoverInit (CK_SESSION_HANDLE handle,
-                          CK_MECHANISM_PTR mechanism,
-                          CK_OBJECT_HANDLE key)
-{
-	return proxy_C_SignRecoverInit (&global.virt.funcs, handle, mechanism, key);
-}
-
-static CK_RV
-module_C_SignRecover (CK_SESSION_HANDLE handle,
-                      CK_BYTE_PTR data,
-                      CK_ULONG data_len,
-                      CK_BYTE_PTR signature,
-                      CK_ULONG_PTR signature_len)
-{
-	return proxy_C_SignRecover (&global.virt.funcs, handle, data, data_len,
-	                            signature, signature_len);
-}
-
-static CK_RV
-module_C_VerifyInit (CK_SESSION_HANDLE handle,
-                     CK_MECHANISM_PTR mechanism,
-                     CK_OBJECT_HANDLE key)
-{
-	return proxy_C_VerifyInit (&global.virt.funcs, handle, mechanism, key);
-}
-
-static CK_RV
-module_C_Verify (CK_SESSION_HANDLE handle,
-                 CK_BYTE_PTR data,
-                 CK_ULONG data_len,
-                 CK_BYTE_PTR signature,
-                 CK_ULONG signature_len)
-{
-	return proxy_C_Verify (&global.virt.funcs, handle, data, data_len, signature,
-	                       signature_len);
-}
-
-static CK_RV
-module_C_VerifyUpdate (CK_SESSION_HANDLE handle,
-                       CK_BYTE_PTR part,
-                       CK_ULONG part_len)
-{
-	return proxy_C_VerifyUpdate (&global.virt.funcs, handle, part, part_len);
-}
-
-static CK_RV
-module_C_VerifyFinal (CK_SESSION_HANDLE handle,
-                      CK_BYTE_PTR signature,
-                      CK_ULONG signature_len)
-{
-	return proxy_C_VerifyFinal (&global.virt.funcs, handle, signature, signature_len);
-}
-
-static CK_RV
-module_C_VerifyRecoverInit (CK_SESSION_HANDLE handle,
-                            CK_MECHANISM_PTR mechanism,
-                            CK_OBJECT_HANDLE key)
-{
-	return proxy_C_VerifyRecoverInit (&global.virt.funcs, handle, mechanism, key);
-}
-
-static CK_RV
-module_C_VerifyRecover (CK_SESSION_HANDLE handle,
-                        CK_BYTE_PTR signature,
-                        CK_ULONG signature_len,
-                        CK_BYTE_PTR data,
-                        CK_ULONG_PTR data_len)
-{
-	return proxy_C_VerifyRecover (&global.virt.funcs, handle, signature, signature_len,
-	                              data, data_len);
-}
-
-static CK_RV
-module_C_DigestEncryptUpdate (CK_SESSION_HANDLE handle,
-                              CK_BYTE_PTR part,
-                              CK_ULONG part_len,
-                              CK_BYTE_PTR enc_part,
-                              CK_ULONG_PTR enc_part_len)
-{
-	return proxy_C_DigestEncryptUpdate (&global.virt.funcs, handle, part, part_len,
-	                                    enc_part, enc_part_len);
-}
-
-static CK_RV
-module_C_DecryptDigestUpdate (CK_SESSION_HANDLE handle,
-                              CK_BYTE_PTR enc_part,
-                              CK_ULONG enc_part_len,
-                              CK_BYTE_PTR part,
-                              CK_ULONG_PTR part_len)
-{
-	return proxy_C_DecryptDigestUpdate (&global.virt.funcs, handle, enc_part,
-	                                    enc_part_len, part, part_len);
-}
-
-static CK_RV
-module_C_SignEncryptUpdate (CK_SESSION_HANDLE handle,
-                            CK_BYTE_PTR part,
-                            CK_ULONG part_len,
-                            CK_BYTE_PTR enc_part,
-                            CK_ULONG_PTR enc_part_len)
-{
-	return proxy_C_SignEncryptUpdate (&global.virt.funcs, handle, part, part_len,
-	                                  enc_part, enc_part_len);
-}
-
-static CK_RV
-module_C_DecryptVerifyUpdate (CK_SESSION_HANDLE handle,
-                              CK_BYTE_PTR enc_part,
-                              CK_ULONG enc_part_len,
-                              CK_BYTE_PTR part,
-                              CK_ULONG_PTR part_len)
-{
-	return proxy_C_DecryptVerifyUpdate (&global.virt.funcs, handle, enc_part,
-	                                    enc_part_len, part, part_len);
-}
-
-static CK_RV
-module_C_GenerateKey (CK_SESSION_HANDLE handle,
-                      CK_MECHANISM_PTR mechanism,
-                      CK_ATTRIBUTE_PTR template,
-                      CK_ULONG count,
-                      CK_OBJECT_HANDLE_PTR key)
-{
-	return proxy_C_GenerateKey (&global.virt.funcs, handle, mechanism, template, count,
-	                            key);
-}
-
-static CK_RV
-module_C_GenerateKeyPair (CK_SESSION_HANDLE handle,
-                          CK_MECHANISM_PTR mechanism,
-                          CK_ATTRIBUTE_PTR pub_template,
-                          CK_ULONG pub_count,
-                          CK_ATTRIBUTE_PTR priv_template,
-                          CK_ULONG priv_count,
-                          CK_OBJECT_HANDLE_PTR pub_key,
-                          CK_OBJECT_HANDLE_PTR priv_key)
-{
-	return proxy_C_GenerateKeyPair (&global.virt.funcs, handle, mechanism, pub_template,
-	                                pub_count, priv_template, priv_count,
-	                                pub_key, priv_key);
-}
-
-static CK_RV
-module_C_WrapKey (CK_SESSION_HANDLE handle,
-                  CK_MECHANISM_PTR mechanism,
-                  CK_OBJECT_HANDLE wrapping_key,
-                  CK_OBJECT_HANDLE key,
-                  CK_BYTE_PTR wrapped_key,
-                  CK_ULONG_PTR wrapped_key_len)
-{
-	return proxy_C_WrapKey (&global.virt.funcs, handle, mechanism, wrapping_key,
-	                        key, wrapped_key, wrapped_key_len);
-}
-
-static CK_RV
-module_C_UnwrapKey (CK_SESSION_HANDLE handle,
-                    CK_MECHANISM_PTR mechanism,
-                    CK_OBJECT_HANDLE unwrapping_key,
-                    CK_BYTE_PTR wrapped_key,
-                    CK_ULONG wrapped_key_len,
-                    CK_ATTRIBUTE_PTR template,
-                    CK_ULONG count,
-                    CK_OBJECT_HANDLE_PTR key)
-{
-	return proxy_C_UnwrapKey (&global.virt.funcs, handle, mechanism, unwrapping_key,
-	                          wrapped_key, wrapped_key_len, template,
-	                          count, key);
-}
-
-static CK_RV
-module_C_DeriveKey (CK_SESSION_HANDLE handle,
-                    CK_MECHANISM_PTR mechanism,
-                    CK_OBJECT_HANDLE base_key,
-                    CK_ATTRIBUTE_PTR template,
-                    CK_ULONG count,
-                    CK_OBJECT_HANDLE_PTR key)
-{
-	return proxy_C_DeriveKey (&global.virt.funcs, handle, mechanism, base_key,
-	                          template, count, key);
-}
-
-static CK_RV
-module_C_SeedRandom (CK_SESSION_HANDLE handle,
-                     CK_BYTE_PTR seed,
-                     CK_ULONG seed_len)
-{
-	return proxy_C_SeedRandom (&global.virt.funcs, handle, seed, seed_len);
-}
-
-static CK_RV
-module_C_GenerateRandom (CK_SESSION_HANDLE handle,
-                         CK_BYTE_PTR random_data,
-                         CK_ULONG random_len)
-{
-	return proxy_C_GenerateRandom (&global.virt.funcs, handle, random_data, random_len);
-}
-
 /* --------------------------------------------------------------------
  * MODULE ENTRY POINT
  */
 
-static CK_FUNCTION_LIST module_functions = {
-	{ CRYPTOKI_VERSION_MAJOR, CRYPTOKI_VERSION_MINOR },
-	module_C_Initialize,
-	module_C_Finalize,
-	module_C_GetInfo,
-	module_C_GetFunctionList,
-	module_C_GetSlotList,
-	module_C_GetSlotInfo,
-	module_C_GetTokenInfo,
-	module_C_GetMechanismList,
-	module_C_GetMechanismInfo,
-	module_C_InitToken,
-	module_C_InitPIN,
-	module_C_SetPIN,
-	module_C_OpenSession,
-	module_C_CloseSession,
-	module_C_CloseAllSessions,
-	module_C_GetSessionInfo,
-	module_C_GetOperationState,
-	module_C_SetOperationState,
-	module_C_Login,
-	module_C_Logout,
-	module_C_CreateObject,
-	module_C_CopyObject,
-	module_C_DestroyObject,
-	module_C_GetObjectSize,
-	module_C_GetAttributeValue,
-	module_C_SetAttributeValue,
-	module_C_FindObjectsInit,
-	module_C_FindObjects,
-	module_C_FindObjectsFinal,
-	module_C_EncryptInit,
-	module_C_Encrypt,
-	module_C_EncryptUpdate,
-	module_C_EncryptFinal,
-	module_C_DecryptInit,
-	module_C_Decrypt,
-	module_C_DecryptUpdate,
-	module_C_DecryptFinal,
-	module_C_DigestInit,
-	module_C_Digest,
-	module_C_DigestUpdate,
-	module_C_DigestKey,
-	module_C_DigestFinal,
-	module_C_SignInit,
-	module_C_Sign,
-	module_C_SignUpdate,
-	module_C_SignFinal,
-	module_C_SignRecoverInit,
-	module_C_SignRecover,
-	module_C_VerifyInit,
-	module_C_Verify,
-	module_C_VerifyUpdate,
-	module_C_VerifyFinal,
-	module_C_VerifyRecoverInit,
-	module_C_VerifyRecover,
-	module_C_DigestEncryptUpdate,
-	module_C_DecryptDigestUpdate,
-	module_C_SignEncryptUpdate,
-	module_C_DecryptVerifyUpdate,
-	module_C_GenerateKey,
-	module_C_GenerateKeyPair,
-	module_C_WrapKey,
-	module_C_UnwrapKey,
-	module_C_DeriveKey,
-	module_C_SeedRandom,
-	module_C_GenerateRandom,
-	module_C_GetFunctionStatus,
-	module_C_CancelFunction,
-	module_C_WaitForSlotEvent
-};
-
 static CK_X_FUNCTION_LIST proxy_functions = {
 	{ CRYPTOKI_VERSION_MAJOR, CRYPTOKI_VERSION_MINOR  },
 	proxy_C_Initialize,
@@ -2349,24 +1664,15 @@ CK_RV
 C_GetFunctionList (CK_FUNCTION_LIST_PTR_PTR list)
 {
 	CK_FUNCTION_LIST_PTR module = NULL;
-	CK_FUNCTION_LIST **loaded;
+	CK_FUNCTION_LIST **loaded = NULL;
 	State *state;
 	CK_RV rv = CKR_OK;
 
 	p11_library_init_once ();
 	p11_lock ();
 
-	if (all_modules == NULL) {
-		/* WARNING: Reentrancy can occur here */
-		rv = p11_modules_load_inlock_reentrant (0, &loaded);
-		if (rv == CKR_OK) {
-			if (all_modules == NULL)
-				all_modules = loaded;
-			else
-				p11_modules_release_inlock_reentrant (loaded);
-		}
-	}
-
+	/* WARNING: Reentrancy can occur here */
+	rv = p11_modules_load_inlock_reentrant (P11_KIT_MODULE_LOADED_FROM_PROXY, &loaded);
 	if (rv == CKR_OK) {
 		state = calloc (1, sizeof (State));
 		if (!state) {
@@ -2375,6 +1681,8 @@ C_GetFunctionList (CK_FUNCTION_LIST_PTR_PTR list)
 		} else {
 			p11_virtual_init (&state->virt, &proxy_functions, state, NULL);
 			state->last_handle = FIRST_HANDLE;
+			state->loaded = loaded;
+			loaded = NULL;
 
 			module = p11_virtual_wrap (&state->virt, free);
 			if (module == NULL) {
@@ -2389,14 +1697,14 @@ C_GetFunctionList (CK_FUNCTION_LIST_PTR_PTR list)
 	}
 
 	if (rv == CKR_OK) {
-		if (module == NULL)
-			module = &module_functions;
-
 		/* We use this as a check below */
 		module->C_WaitForSlotEvent = module_C_WaitForSlotEvent;
 		*list = module;
 	}
 
+	if (loaded)
+		p11_kit_modules_release (loaded);
+
 	p11_unlock ();
 
 	return rv;
@@ -2413,11 +1721,7 @@ p11_proxy_module_cleanup (void)
 	for (; state != NULL; state = next) {
 		next = state->next;
 		p11_virtual_unwrap (state->wrapped);
-	}
-
-	if (all_modules) {
-		p11_kit_modules_release (all_modules);
-		all_modules = NULL;
+		p11_kit_modules_release (state->loaded);
 	}
 }
 
@@ -2426,3 +1730,41 @@ p11_proxy_module_check (CK_FUNCTION_LIST_PTR module)
 {
 	return (module->C_WaitForSlotEvent == module_C_WaitForSlotEvent);
 }
+
+static void
+proxy_module_free (p11_virtual *virt)
+{
+	State *state = (State *)virt;
+
+	p11_virtual_unwrap (state->wrapped);
+	p11_kit_modules_release (state->loaded);
+	free (state);
+}
+
+CK_RV
+p11_proxy_module_create (CK_FUNCTION_LIST_PTR *module,
+			 CK_FUNCTION_LIST_PTR *modules)
+{
+	State *state;
+	CK_RV rv = CKR_OK;
+
+	assert (module != NULL);
+	assert (modules != NULL);
+
+	state = calloc (1, sizeof (State));
+	if (!state)
+		return CKR_HOST_MEMORY;
+
+	p11_virtual_init (&state->virt, &proxy_functions, state, NULL);
+	state->last_handle = FIRST_HANDLE;
+	state->loaded = modules_dup (modules);
+	state->wrapped = p11_virtual_wrap (&state->virt, (p11_destroyer)proxy_module_free);
+	if (state->wrapped == NULL) {
+		proxy_module_free (&state->virt);
+		return CKR_GENERAL_ERROR;
+	}
+
+	*module = state->wrapped;
+
+	return rv;
+}
