@@ -45,6 +45,7 @@
 #include "library.h"
 #include "message.h"
 #include "module.h"
+#include "oid.h"
 #include "parser.h"
 #include "path.h"
 #include "pkcs11.h"
@@ -77,6 +78,8 @@ typedef struct _FindObjects {
 	CK_ATTRIBUTE *match;
 	CK_OBJECT_HANDLE *snapshot;
 	CK_ULONG iterator;
+	CK_ATTRIBUTE *public_key;
+	p11_dict *extensions;
 } FindObjects;
 
 static CK_FUNCTION_LIST sys_function_list;
@@ -87,6 +90,7 @@ find_objects_free (void *data)
 	FindObjects *find = data;
 	p11_attrs_free (find->match);
 	free (find->snapshot);
+	p11_dict_free (find->extensions);
 	free (find);
 }
 
@@ -194,10 +198,11 @@ create_tokens_inlock (p11_array *tokens,
 	struct {
 		const char *prefix;
 		const char *label;
+		int flags;
 	} labels[] = {
-		{ "~/", "User Trust" },
-		{ DATA_DIR, "Default Trust" },
-		{ SYSCONFDIR, "System Trust" },
+		{ "~/", "User Trust", P11_TOKEN_FLAG_NONE },
+		{ P11_DEFAULT_TRUST_PREFIX, "Default Trust", P11_TOKEN_FLAG_WRITE_PROTECTED },
+		{ P11_SYSTEM_TRUST_PREFIX, "System Trust", P11_TOKEN_FLAG_NONE },
 		{ NULL },
 	};
 
@@ -206,6 +211,7 @@ create_tokens_inlock (p11_array *tokens,
 	CK_SLOT_ID slot;
 	const char *path;
 	const char *label;
+	int flags;
 	char *alloc;
 	char *remaining;
 	char *base;
@@ -232,12 +238,14 @@ create_tokens_inlock (p11_array *tokens,
 			slot = BASE_SLOT_ID + tokens->num;
 
 			label = NULL;
+			flags = P11_TOKEN_FLAG_NONE;
 			base = NULL;
 
 			/* Claim the various labels based on prefix */
 			for (i = 0; label == NULL && labels[i].prefix != NULL; i++) {
 				if (strncmp (path, labels[i].prefix, strlen (labels[i].prefix)) == 0) {
 					label = labels[i].label;
+					flags = labels[i].flags;
 					labels[i].label = NULL;
 				}
 			}
@@ -248,7 +256,7 @@ create_tokens_inlock (p11_array *tokens,
 				return_val_if_fail (base != NULL, false);
 			}
 
-			token = p11_token_new (slot, path, label);
+			token = p11_token_new (slot, path, label, flags);
 			return_val_if_fail (token != NULL, false);
 
 			if (!p11_array_push (tokens, token))
@@ -436,8 +444,8 @@ sys_C_GetInfo (CK_INFO_PTR info)
 		info->libraryVersion.major = PACKAGE_MAJOR;
 		info->libraryVersion.minor = PACKAGE_MINOR;
 		info->flags = 0;
-		strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
-		strncpy ((char*)info->libraryDescription, LIBRARY_DESCRIPTION, 32);
+		memcpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
+		memcpy ((char*)info->libraryDescription, LIBRARY_DESCRIPTION, 32);
 	}
 
 	p11_debug ("out: 0x%lx", rv);
@@ -519,7 +527,7 @@ sys_C_GetSlotInfo (CK_SLOT_ID id,
 		info->hardwareVersion.major = PACKAGE_MAJOR;
 		info->hardwareVersion.minor = PACKAGE_MINOR;
 		info->flags = CKF_TOKEN_PRESENT;
-		strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
+		memcpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
 
 		/* If too long, copy the first 64 characters into buffer */
 		path = p11_token_get_path (token);
@@ -559,9 +567,9 @@ sys_C_GetTokenInfo (CK_SLOT_ID id,
 		info->hardwareVersion.major = PACKAGE_MAJOR;
 		info->hardwareVersion.minor = PACKAGE_MINOR;
 		info->flags = CKF_TOKEN_INITIALIZED;
-		strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
-		strncpy ((char*)info->model, TOKEN_MODEL, 16);
-		strncpy ((char*)info->serialNumber, TOKEN_SERIAL_NUMBER, 16);
+		memcpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
+		memcpy ((char*)info->model, TOKEN_MODEL, 16);
+		memcpy ((char*)info->serialNumber, TOKEN_SERIAL_NUMBER, 16);
 		info->ulMaxSessionCount = CK_EFFECTIVELY_INFINITE;
 		info->ulSessionCount = CK_UNAVAILABLE_INFORMATION;
 		info->ulMaxRwSessionCount = 0;
@@ -836,6 +844,17 @@ sys_C_Login (CK_SESSION_HANDLE handle,
 	p11_lock ();
 
 		rv = lookup_session (handle, NULL);
+		/* Since the trust module is designed as a replacement
+		 * of nssckbi, it works as a general access device as
+		 * described in the table 1.1 of:
+		 * <https://developer.mozilla.org/en-US/docs/Mozilla/Projects/NSS/PKCS11_Implement>.
+		 *
+		 * That means that the tokens provided by this module
+		 * shall be accessed without login, and if the caller
+		 * tries to login, the attempt should fail with an
+		 * explicit error (otherwise, the caller cannot
+		 * distinguish the user's login status, see also
+		 * C_Logout below). */
 		if (rv == CKR_OK)
 			rv = CKR_USER_TYPE_INVALID;
 
@@ -1147,6 +1166,7 @@ sys_C_FindObjectsInit (CK_SESSION_HANDLE handle,
 	char *string;
 	CK_RV rv;
 	int n = 0;
+	CK_OBJECT_CLASS klass;
 
 	if (p11_debugging) {
 		string = p11_attrs_to_string (template, count);
@@ -1190,6 +1210,14 @@ sys_C_FindObjectsInit (CK_SESSION_HANDLE handle,
 				find->iterator = 0;
 				find->snapshot = p11_index_snapshot (indices[0], indices[1], template, count);
 				warn_if_fail (find->snapshot != NULL);
+
+				if (p11_attrs_find_ulong (find->match, CKA_CLASS, &klass) &&
+				    klass == CKO_X_CERTIFICATE_EXTENSION) {
+					find->public_key = p11_attrs_find (find->match, CKA_PUBLIC_KEY_INFO);
+					find->extensions = p11_dict_new (p11_oid_hash,
+									 p11_oid_equal,
+									 free, NULL);
+				}
 			}
 
 			if (!find || !find->snapshot || !find->match)
@@ -1243,10 +1271,10 @@ match_for_broken_nss_serial_number_lookups (CK_ATTRIBUTE *attr,
 
 static bool
 find_objects_match (CK_ATTRIBUTE *attrs,
-                    CK_ATTRIBUTE *match)
+                    FindObjects *find)
 {
 	CK_OBJECT_CLASS klass;
-	CK_ATTRIBUTE *attr;
+	CK_ATTRIBUTE *attr, *match = find->match;
 
 	for (; !p11_attrs_terminator (match); match++) {
 		attr = p11_attrs_find ((CK_ATTRIBUTE *)attrs, match->type);
@@ -1274,6 +1302,29 @@ find_objects_match (CK_ATTRIBUTE *attrs,
 		return false;
 	}
 
+	/*
+	 * WORKAROUND: We keep all objects in the database, while PKIX
+	 * doesn't allow multiple extensions identified by the same
+	 * OID can be attached to a certificate.  Check any duplicate
+	 * and only return the first matching object.
+	 */
+	if (find->public_key &&
+	    p11_attrs_find_ulong (attrs, CKA_CLASS, &klass) &&
+	    klass == CKO_X_CERTIFICATE_EXTENSION) {
+		CK_ATTRIBUTE *oid = p11_attrs_find (attrs, CKA_OBJECT_ID);
+		if (oid) {
+			void *value;
+			if (p11_oid_simple (oid->pValue, oid->ulValueLen) &&
+			    p11_dict_get (find->extensions, oid->pValue)) {
+				p11_debug ("duplicate extension object");
+				return false;
+			}
+			value = memdup (oid->pValue, oid->ulValueLen);
+			return_val_if_fail (value != NULL, false);
+			p11_dict_set (find->extensions, value, value);
+		}
+	}
+
 	return true;
 }
 
@@ -1317,7 +1368,7 @@ sys_C_FindObjects (CK_SESSION_HANDLE handle,
 				if (attrs == NULL)
 					continue;
 
-				if (find_objects_match (attrs, find->match)) {
+				if (find_objects_match (attrs, find)) {
 					objects[matched] = object;
 					matched++;
 				}
@@ -1794,7 +1845,7 @@ __attribute__((constructor))
 void
 p11_trust_module_init (void)
 {
-	p11_library_init_once ();
+	p11_library_init ();
 }
 
 #ifdef __GNUC__
